import pandas as pd
import numpy as np
import zarr
import os
import json

class BSCCM:

    def __init__(self, data_root, cache_index=False):
        """
        
        data_root: path to the top-level BSCCM directory
        cache_index: load the full index into memory. Set to true for increased performance at the expense of memory usage
        """
        print('Opening BSCCM (this may take a few seconds)...\r', end='')
        self.data_root = data_root
        self.zarr_dataset = zarr.open(data_root + 'BSCCM_images.zarr', 'r')
        self.index_dataframe = pd.read_csv(data_root + 'BSCCM_index.csv', low_memory=not cache_index, index_col='global_index')
        self.global_metadata = json.loads(open(data_root + 'BSCCM_global_metadata.json').read())
        self.size = len(self.index_dataframe)
        self.fluor_channel_names = self.global_metadata['fluorescence']['channel_names']
        self.led_array_channel_names = self.global_metadata['led_array']['channel_names']
        if 'BSCCM_surface_markers.csv' in os.listdir(data_root):
            self.surface_marker_dataframe = pd.read_csv(data_root + 'BSCCM_surface_markers.csv', index_col='global_index')
        if 'BSCCM_backgrounds.zarr' in os.listdir(data_root):
            self.backgrounds_and_shading = zarr.open(data_root + 'BSCCM_backgrounds.zarr', 'r')
        print('BSCCM Opened                                     ')

    
    def read_image(self, index, contrast_type, channel=None, copy=False, convert_histology_rgb32=True):
        """
        
        TODO: add a note about how histology is translated on the fly to RGB32
        
        contrast_type: 'led_array', 'fluor', 'dpc', 'histology'
        """
        
        entry = self.index_dataframe.loc[index]
        base_path = entry['data_path'] + '/'
        if contrast_type == 'led_array':
            base_path += contrast_type
            channel_index = self.led_array_channel_names.index(channel)
        elif contrast_type == 'fluor':
            base_path += contrast_type
            channel_index = self.fluor_channel_names.index(channel)
        elif contrast_type == 'dpc':
            base_path += contrast_type
            channel_index = 0
        elif contrast_type == 'histology':
            base_path += contrast_type
            channel_index = None
        else:
            raise Exception('unrecognized contrast_type')
        
        image = self.zarr_dataset[base_path + '/cell_{}'.format(index)]
        if channel_index is not None:
            image = image[channel_index]
        
        if contrast_type == 'histology' and convert_histology_rgb32:
            image = np.array(image) / (2 ** 12)
            
        if copy:
            return np.array(image)
        return image
        
    def get_indices(self, batch=None, slide_replicate=None, antibodies=None, 
                    has_matched_histology=False, shuffle=False, seed=None):
        sub_data_frame = self.index_dataframe
        if batch is not None:
            sub_data_frame = sub_data_frame[sub_data_frame.batch == batch]
        if slide_replicate is not None:
            sub_data_frame = sub_data_frame[sub_data_frame.slide_replicate == slide_replicate]
        if antibodies is not None:
            if type(antibodies) == list or type(antibodies) == tuple:
                sub_data_frame = sub_data_frame[sub_data_frame.antibodies.isin(antibodies)]    
            else:
                sub_data_frame = sub_data_frame[sub_data_frame.antibodies == antibodies]
        if has_matched_histology:
            sub_data_frame = sub_data_frame[sub_data_frame.has_matched_histology_cell]
        
        indices = sub_data_frame.index.to_numpy()
        if shuffle:
            if seed is not None:
                np.random.seed(seed)
            np.random.shuffle(indices)
        return indices
    
    def get_corrected_fluor(self, indices):
        names = [
            'Fluor_690-_shading_corrected',   
            'Fluor_627-673_shading_corrected', 
            'Fluor_585-625_shading_corrected',             
            'Fluor_550-570_shading_corrected', 
            'Fluor_500-550_shading_corrected', 
            'Fluor_426-446_shading_corrected',
               ]
        return self.surface_marker_dataframe.loc[indices][names].to_numpy()
    
    def get_surface_marker_data(self, indices):
        four_spectra_model_names = [
           'CD123/HLA-DR/CD14_full_model_unmixed',
           'CD3/CD19/CD56_full_model_unmixed', 
           'CD45_full_model_unmixed',
           'CD16_full_model_unmixed']
            
        two_spectra_model_names = [
                           'CD45_single_antibody_model_unmixed',
                           'autofluor_single_antibody_model_unmixed',
                           'CD123_single_antibody_model_unmixed',
                           'CD19_single_antibody_model_unmixed',
                           'CD56_single_antibody_model_unmixed',
                           'CD14_single_antibody_model_unmixed',
                           'CD16_single_antibody_model_unmixed',
                           'HLA-DR_single_antibody_model_unmixed',
                           'CD3_single_antibody_model_unmixed',
                          ]
        
        four_spectra_data = self.surface_marker_dataframe.loc[indices][four_spectra_model_names].to_numpy()
        two_spectra_data = self.surface_marker_dataframe.loc[indices][two_spectra_model_names].to_numpy()
        return two_spectra_model_names, two_spectra_data, four_spectra_model_names, four_spectra_data
    
    def get_background(self, index, channel='Brightfield', percentile=50, dim=128):
        """
        Return a pixel-wise percentile over many images from the cell's location in the original data,
        specific to each physical slide
        
        contrast_type: 'led_array', 'fluor', 'dpc', 'histology'
        percentile: 5, 10, 20, 40, 50 (median) 
        """
        if percentile not in [5, 10, 20, 40, 50]:
            raise Exception('percentile must be one of: [5, 10, 20, 40, 50]')
        
        image = self.backgrounds_and_shading[channel]['{}_percentile'.format(percentile)]
  
        x, y = self.index_dataframe.loc[index].position_in_fov_x_pix, self.index_dataframe.loc[index].position_in_fov_y_pix
        cropped = image[..., int(y) - dim // 2: int(y) + dim // 2, int(x) - dim // 2: int(x) + dim // 2]
        return cropped
    
    


"""
Several helper functions for retrieving information about and plotting
the calibration of the LED array
"""
import argparse
import pandas as pd
from pathlib import Path
import numpy as np
import json
import os
import matplotlib.pyplot as plt

def cartesian_to_na(point_list_cart, z_offset=8):
    """functions for calcuating the NA of an LED on the quasi-dome based on it's index for the quasi-dome illuminator
    converts a list of cartesian points to numerical aperture (NA)

    Args:
        point_list_cart: List of (x,y,z) positions relative to the sample (origin)
        z_offset : Optional, offset of LED array in z, mm

    Returns:
        A 2D numpy array where the first dimension is the number of LEDs loaded and the second is (Na_x, NA_y)
    """
    yz = np.sqrt(point_list_cart[:, 1] ** 2 + (point_list_cart[:, 2] + z_offset) ** 2)
    xz = np.sqrt(point_list_cart[:, 0] ** 2 + (point_list_cart[:, 2] + z_offset) ** 2)
    xyz = np.sqrt(point_list_cart[:, 0] ** 2 + point_list_cart[:, 1] ** 2 + (point_list_cart[:, 2] + z_offset) ** 2)

    na_xy = np.zeros((np.size(point_list_cart, 0), 2))
    na_xy[:, 0] = np.sin(np.arctan(point_list_cart[:, 0] / yz))
    na_xy[:, 1] = np.sin(np.arctan(point_list_cart[:, 1] / xz))

    xy = np.sqrt(point_list_cart[:, 0] ** 2 + point_list_cart[:, 1] ** 2 )
    na = np.sin(np.arctan(xy / (point_list_cart[:, 2] + z_offset)))
#     na = np.sin(np.arctan(np.sqrt(point_list_cart[:, 0] ** 2 
#                                   + point_list_cart[:, 1] ** 2 ) / xyz))
    
    return na_xy, na

def load_led_positions_from_json(file_name, z_offset=8):
    """Function which loads LED positions from a json file
    Args:
        fileName: Location of file to load
        zOffset : Optional, offset of LED array in z, mm
        micro : 'TE300B' or 'TE300A'
    Returns:
        A 2D numpy array where the first dimension is the number of LEDs loaded and the second is (x, y, z) in mm
    """
    json_data = open(file_name).read()
    data = json.loads(json_data)

    source_list_cart = np.zeros((len(data['led_list']), 3))
    x = [d['x'] for d in data['led_list']]
    y = [d['y'] for d in data['led_list']]
    z = [d['z'] for d in data['led_list']]

    source_list_cart[:, 0] = x
    source_list_cart[:, 1] = y
    source_list_cart[:, 2] = z

    source_list_naxy, na = cartesian_to_na(source_list_cart, z_offset=z_offset)

    return source_list_naxy, na, source_list_cart

def get_led_na(led_index):
    source_list_na_xy, source_list_na, source_list_cart = load_led_positions_from_json(os.path.dirname(__file__) + '/quasi_dome_design.json')
    return source_list_na[led_index - 1]
#     angles_xy = np.arcsin(np.abs(source_list_na))
#     angle = np.arctan(np.sqrt(np.tan(angles_xy[:, 0])**2 + np.tan(angles_xy[:, 1])**2 ))
#     return np.sin(angle[led_index - 1])

def get_led_na_xy(led_index):
    """
    get na x na y based on 1-based index
    :param led_index:
    :return:
    """
    source_list_na_xy, source_list_na, source_list_cart = load_led_positions_from_json(os.path.dirname(__file__) + '/quasi_dome_design.json')
#     if np.min(led_index) < 1 or np.max(led_index) > source_list_na_xy.shape[0]:
#         raise Exception('LED index out of range')
    return source_list_na_xy[led_index - 1]

def get_led_angle(led_index):
    source_list_na, source_list_cart = load_led_positions_from_json(os.path.dirname(__file__) + '/quasi_dome_design.json')
    angles_xy = np.arcsin(np.abs(source_list_na))
    angle = np.arctan(np.sqrt(np.tan(angles_xy[:, 0])**2 + np.tan(angles_xy[:, 1])**2 ))
    return angle[led_index - 1] / (2*3.14) *360

def plot_led_pattern(led_indices=None, channel_name=None, ax=None, legend=True, size=20):
    """
    Make a plot of the the illumination pattern in NA space
    """
    if channel_name is None and led_indices is None:
        raise Exception('Must supply either channel_name or led_indices')
    if channel_name is not None and led_indices is not None:
        raise Exception('Must supply only channel_name or led_indices')
    if channel_name is not None:
        led_indices = illumination_to_led_indices(channel_name)

    if ax is None:
        ax = plt.gca()
    all_led_list = np.array([get_led_na_xy(led) for led in range(1, 582)])
    ax.scatter(all_led_list[:,0], all_led_list[:,1], size, marker=',', color=[0.15,0.15,0.15], 
                facecolor=None,edgecolor=None, edgecolors=None)
    ax.set_facecolor([0,0,0])
    
    for led_index in led_indices:
        nax, nay = get_led_na_xy(led_index) 
        ax.scatter(nax, nay, size, marker='s', color=[0,1.0,0], 
                    facecolor=None, edgecolor=None, edgecolors=None)
    ax.add_patch(plt.Circle((0, 0), 0.5, alpha=0.25))
    ax.set_xlabel('Numerical aperture (x)')
    ax.set_ylabel('Numerical aperture (y)')
    if legend:
        plt.legend(['Brightfield LEDs','LED off','LED on'])


def illumination_to_led_indices(channel):
    """
    Compute a channel name to a list of LED indices. LED indices are 1-based
    """
    if channel == 'LED119':
        return [119]
    
    if 'led'in channel:
        return [int(channel.split('_')[-1])]
    
    def include_fn(na, na_xy):
        mask = np.array(na.size * [True])
        if channel == 'Brightfield':
            mask[na > 0.5] = False
        
        if 'Top' in channel:
            mask[na_xy[:, 1] < 0] = False
        elif 'Bottom' in channel:
            mask[na_xy[:, 1] > 0] = False
        elif 'Left' in channel:
            mask[na_xy[:, 0] < 0] = False
        elif 'Right' in channel:
            mask[na_xy[:, 0] > 0] = False
            
        if 'DPC' in channel:
            mask[na > 0.5] = False
            mask[na < 0.4] = False
            
        if 'DF' in channel:
            na_min = float(channel.split('_')[1]) / 100
            mask[na > na_min + 0.05] = False
            mask[na < na_min] = False
        return mask
    
    led_indices = np.arange(1, 582)
    na = get_led_na(led_indices)
    na_xy = get_led_na_xy(led_indices)
    mask = include_fn(na, na_xy)
    return led_indices[mask]
    